/*
 * UniformParametricGenerator.java
 *
 * Created on August 3, 2006, 2:12 AM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package sddm.data;

import java.util.Vector;

/**
 * This is a ParametricGenerator where the time is uniformly spaced between elements
 * @author Erik J. Tollerud
 */
public class UniformParametricGenerator implements ParametricGenerator {
    private float deltat;
    private float x0;
    private float y0;
    private float z0;
    private float w0;
    private float t0;
    private Generator gen;
    
    /** Creates a new instance of UniformParametricGenerator */
    public UniformParametricGenerator(Generator gen) {
        this.setDeltat(1.0f);
        this.setX0(0.0f);
        this.setY0(0.0f);
        this.setZ0(0.0f);
        this.setW0(1.0f);
        this.setT0(0.0f);
        this.gen = gen;
    }
    
    public UniformParametricGenerator(float deltat, Generator gen) {
        this.setDeltat(deltat);
        this.setX0(0.0f);
        this.setY0(0.0f);
        this.setZ0(0.0f);
        this.setW0(1.0f);
        this.setT0(0.0f);
        this.gen = gen;
    }
    
    public UniformParametricGenerator(Generator gen, float x0, float y0, float z0) {
        this.setDeltat(1.0f);
        this.setX0(x0);
        this.setY0(y0);
        this.setZ0(z0);
        this.setW0(1.0f);
        this.setT0(0.0f);
        this.gen = gen;
    } 
    
    public UniformParametricGenerator(Generator gen, float x0, float y0, float z0, float w0) {
        this.setDeltat(1.0f);
        this.setX0(x0);
        this.setY0(y0);
        this.setZ0(z0);
        this.setW0(w0);
        this.setT0(0.0f);
        this.gen = gen;
    } 
    
    public UniformParametricGenerator(float deltat, Generator gen, float x0, float y0, float z0, float t0) {
        this.setDeltat(deltat);
        this.setX0(x0);
        this.setY0(y0);
        this.setZ0(z0);
        this.setW0(1.0f);
        this.setT0(t0);
        this.gen = gen;
    }
    
    public UniformParametricGenerator(float deltat, Generator gen, float x0, float y0, float z0, float w0, float t0) {
        this.setDeltat(deltat);
        this.setX0(x0);
        this.setY0(y0);
        this.setZ0(z0);
        this.setW0(w0);
        this.setT0(t0);
        this.gen = gen;
    } 

    public float[] nextIteration(Vector<Float> time, PointList wholeList) {
        float[] res = getGen().nextIteration(time.lastElement(),getDeltat(), wholeList);
        time.add(time.lastElement()+getDeltat());
            
        return res;
    }

    public float[] initialZs() {
        float[] res = new float[1];
        res[0] = getZ0();
        return res;
    }

    public float[] initialYs() {
        float[] res = new float[1];
        res[0] = getY0();
        return res;
    }

    public float[] initialXs() {
        float[] res = new float[1];
        res[0] = getX0();
        return res;
    }

    public float[] initialWs() {
        float[] res = new float[1];
        res[0] = getW0();
        return res;
    }

    public float[] initialTs() {
        float[] res = new float[1];
        res[0] = getT0();
        return res;
    }

    public float getDeltat() {
        return deltat;
    }

    public void setDeltat(float deltat) {
        this.deltat = deltat;
    }

    public float getX0() {
        return x0;
    }

    public void setX0(float x0) {
        this.x0 = x0;
    }

    public float getY0() {
        return y0;
    }

    public void setY0(float y0) {
        this.y0 = y0;
    }

    public float getZ0() {
        return z0;
    }

    public void setZ0(float z0) {
        this.z0 = z0;
    }

    public float getW0() {
        return w0;
    }

    public void setW0(float w0) {
        this.w0 = w0;
    }

    public float getT0() {
        return t0;
    }

    public void setT0(float t0) {
        this.t0 = t0;
    }

    public Generator getGen() {
        return gen;
    }
    
    public String toString() {
        this.gen.toString();
        return "Uniform Parametric Generator with x0="+this.x0+", y0="+this.y0+", z0="+this.z0+ ", w0="+this.w0+", t0="+this.t0+ ", deltat="+this.deltat+" and generator "+gen.toString();
    }
    
    public static interface Generator {
        public float[] nextIteration(float lastTime, float deltat, PointList wholeList);
    }
}
