/*
 * ParametricDiffEqTypePanel.java
 *
 * Created on August 3, 2006, 11:41 AM
 */

package sddm.modules;

import sddm.*;
import sddm.data.*;
import java.util.*;
import java.util.regex.*;

/**
 *
 * @author  Erik J. Tollerud
 */
public class ParametricDiffEqTypePanel extends ParametricModuleSetupPanel.TypePanel {
    ParametricModuleSetupPanel parent;
    
    /** Creates new form ParametricDiffEqTypePanel */
    //TODO:implement Runge-Kudda method
    public ParametricDiffEqTypePanel(ParametricModuleSetupPanel sp) {
        super();
        this.parent = sp;
        initComponents();
    }
    
    public ParametricDiffEqTypePanel(ParametricModuleSetupPanel sp, ParametricSystem initialSystem) {
        this(sp);
        if (initialSystem != null) {
            if (initialSystem.getGen() instanceof UniformParametricGenerator) {
                UniformParametricGenerator upg = (UniformParametricGenerator)initialSystem.getGen();
                UniformParametricGenerator.Generator gen = upg.getGen();
                if (gen instanceof DiffEqGenerator) {
                    DiffEqGenerator deg = (DiffEqGenerator)gen;
                    
                    this.dxField.setText(deg.xe.toString());
                    this.dyField.setText(deg.ye.toString());
                    this.dzField.setText(deg.ze.toString());
                    if (deg.wDefined)
                        this.dwField.setText(deg.we.toString());
                    
                    this.deltaTField.setValue(upg.getDeltat());
                    this.x0Field.setValue(upg.getX0());
                    this.y0Field.setValue(upg.getY0());
                    this.z0Field.setValue(upg.getZ0());
                    this.wCheckBox.setSelected(deg.wDefined);
                    
                    this.itersField.setValue(initialSystem.getCurrentIteration());
                }
            }
        }
    }
    
    public sddm.data.ParametricSystem makeSystem() {
        float x0 = ((Number)this.x0Field.getValue()).floatValue();
        float y0 = ((Number)this.y0Field.getValue()).floatValue();
        float z0 = ((Number)this.z0Field.getValue()).floatValue();
        float t0 = ((Number)this.t0Field.getValue()).floatValue();
        
        float deltat = ((Number)this.deltaTField.getValue()).floatValue();
        int iters = ((Number)this.itersField.getValue()).intValue();
        
        String dxStr = this.dxField.getText();
        String dyStr = this.dyField.getText();
        String dzStr = this.dzField.getText();
        String dwStr = null;
        if (this.wCheckBox.isSelected())
            dwStr = this.dwField.getText();
        
        UniformParametricGenerator gen = new UniformParametricGenerator(deltat,new DiffEqGenerator(dxStr,dyStr,dzStr,dwStr),x0,y0,z0,t0);
        ParametricSystem sys = new ParametricSystem(gen);
        try {
            if (iters == 1)
                sys.iterate();
            else
                sys.iterate(iters);
        } catch (SDDMException e) {
            javax.swing.JOptionPane.showMessageDialog(null,e.toString(),"Invalid Number",javax.swing.JOptionPane.ERROR_MESSAGE);
        }
        
        return sys;
    }
    
    /** This method is called from within the constructor to
     * initialize the form.
     * WARNING: Do NOT modify this code. The content of this method is
     * always regenerated by the Form Editor.
     */
    // <editor-fold defaultstate="collapsed" desc=" Generated Code ">//GEN-BEGIN:initComponents
    private void initComponents() {
        itersField = new javax.swing.JFormattedTextField(java.text.NumberFormat.getInstance());
        deltaTField = new javax.swing.JFormattedTextField(java.text.NumberFormat.getInstance());
        jLabel2 = new javax.swing.JLabel();
        jLabel3 = new javax.swing.JLabel();
        jLabel1 = new javax.swing.JLabel();
        jLabel4 = new javax.swing.JLabel();
        jLabel5 = new javax.swing.JLabel();
        dxField = new javax.swing.JTextField();
        dyField = new javax.swing.JTextField();
        dzField = new javax.swing.JTextField();
        jLabel6 = new javax.swing.JLabel();
        dwField = new javax.swing.JTextField();
        wCheckBox = new javax.swing.JCheckBox();
        x0Field = new javax.swing.JFormattedTextField(java.text.NumberFormat.getInstance());
        jLabel7 = new javax.swing.JLabel();
        y0Field = new javax.swing.JFormattedTextField(java.text.NumberFormat.getInstance());
        z0Field = new javax.swing.JFormattedTextField(java.text.NumberFormat.getInstance());
        jLabel8 = new javax.swing.JLabel();
        t0Field = new javax.swing.JFormattedTextField(java.text.NumberFormat.getInstance());

        setPreferredSize(new java.awt.Dimension(253, 170));
        itersField.setText("num");
        itersField.setValue(1000);

        deltaTField.setText("num");
        deltaTField.setValue(0.01f);

        jLabel2.setText("deltat");

        jLabel3.setText(" iterations");

        jLabel1.setText("dx/dt");

        jLabel4.setText("dy/dt");

        jLabel5.setText("dz/dt");

        jLabel6.setText("dw/dt");

        dwField.setEnabled(false);

        wCheckBox.setText("use w?");
        wCheckBox.setBorder(javax.swing.BorderFactory.createEmptyBorder(0, 0, 0, 0));
        wCheckBox.setMargin(new java.awt.Insets(0, 0, 0, 0));
        wCheckBox.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                wCheckBoxActionPerformed(evt);
            }
        });

        x0Field.setText("num");
        x0Field.setValue(1);

        jLabel7.setText("Initial Point:");

        y0Field.setText("num");
        y0Field.setValue(1);

        z0Field.setText("num");
        z0Field.setValue(1);

        jLabel8.setText("t0");

        t0Field.setText("num");
        t0Field.setValue(0.0f);

        org.jdesktop.layout.GroupLayout layout = new org.jdesktop.layout.GroupLayout(this);
        this.setLayout(layout);
        layout.setHorizontalGroup(
            layout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
            .add(layout.createSequentialGroup()
                .addContainerGap()
                .add(layout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
                    .add(layout.createSequentialGroup()
                        .add(layout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
                            .add(layout.createSequentialGroup()
                                .add(jLabel1)
                                .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                                .add(dxField, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, 156, Short.MAX_VALUE))
                            .add(layout.createSequentialGroup()
                                .add(jLabel4)
                                .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                                .add(dyField, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, 156, Short.MAX_VALUE))
                            .add(layout.createSequentialGroup()
                                .add(jLabel6)
                                .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                                .add(dwField, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, 154, Short.MAX_VALUE))
                            .add(layout.createSequentialGroup()
                                .add(jLabel5)
                                .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                                .add(dzField, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, 157, Short.MAX_VALUE))
                            .add(layout.createSequentialGroup()
                                .add(jLabel7)
                                .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                                .add(x0Field, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE)
                                .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                                .add(y0Field, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE)
                                .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                                .add(z0Field, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE)
                                .add(5, 5, 5)))
                        .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                        .add(wCheckBox))
                    .add(layout.createSequentialGroup()
                        .add(jLabel2)
                        .add(6, 6, 6)
                        .add(deltaTField, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE)
                        .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                        .add(jLabel3)
                        .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                        .add(itersField, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE)
                        .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                        .add(jLabel8)
                        .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                        .add(t0Field, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE)
                        .addContainerGap(27, Short.MAX_VALUE))))
        );
        layout.setVerticalGroup(
            layout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
            .add(layout.createSequentialGroup()
                .addContainerGap()
                .add(layout.createParallelGroup(org.jdesktop.layout.GroupLayout.BASELINE)
                    .add(jLabel1)
                    .add(dxField, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE))
                .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                .add(layout.createParallelGroup(org.jdesktop.layout.GroupLayout.BASELINE)
                    .add(jLabel4)
                    .add(dyField, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE))
                .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                .add(layout.createParallelGroup(org.jdesktop.layout.GroupLayout.BASELINE)
                    .add(jLabel5)
                    .add(dzField, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE))
                .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                .add(layout.createParallelGroup(org.jdesktop.layout.GroupLayout.BASELINE)
                    .add(jLabel6)
                    .add(dwField, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE)
                    .add(wCheckBox))
                .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                .add(layout.createParallelGroup(org.jdesktop.layout.GroupLayout.BASELINE)
                    .add(jLabel7)
                    .add(x0Field, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE)
                    .add(y0Field, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE)
                    .add(z0Field, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE))
                .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                .add(layout.createParallelGroup(org.jdesktop.layout.GroupLayout.BASELINE)
                    .add(jLabel2)
                    .add(deltaTField, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE)
                    .add(jLabel3)
                    .add(itersField, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE)
                    .add(jLabel8)
                    .add(t0Field, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE))
                .addContainerGap())
        );
    }// </editor-fold>//GEN-END:initComponents
    
    private void wCheckBoxActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_wCheckBoxActionPerformed
        this.dwField.setEnabled(wCheckBox.isSelected());
    }//GEN-LAST:event_wCheckBoxActionPerformed
    
    
    // Variables declaration - do not modify//GEN-BEGIN:variables
    private javax.swing.JFormattedTextField deltaTField;
    private javax.swing.JTextField dwField;
    private javax.swing.JTextField dxField;
    private javax.swing.JTextField dyField;
    private javax.swing.JTextField dzField;
    private javax.swing.JFormattedTextField itersField;
    private javax.swing.JLabel jLabel1;
    private javax.swing.JLabel jLabel2;
    private javax.swing.JLabel jLabel3;
    private javax.swing.JLabel jLabel4;
    private javax.swing.JLabel jLabel5;
    private javax.swing.JLabel jLabel6;
    private javax.swing.JLabel jLabel7;
    private javax.swing.JLabel jLabel8;
    private javax.swing.JFormattedTextField t0Field;
    private javax.swing.JCheckBox wCheckBox;
    private javax.swing.JFormattedTextField x0Field;
    private javax.swing.JFormattedTextField y0Field;
    private javax.swing.JFormattedTextField z0Field;
    // End of variables declaration//GEN-END:variables
    
    public static class DiffEqGenerator implements sddm.data.UniformParametricGenerator.Generator {
        CompiledStatement xe;
        CompiledStatement ye;
        CompiledStatement ze;
        CompiledStatement we;
        boolean xDefined;
        boolean yDefined;
        boolean zDefined;
        boolean wDefined;
        
        int numOfVars;
        
        float[] holder;
        
        public DiffEqGenerator(String dxeqn, String dyeqn, String dzeqn, String dweqn) {
            xDefined = (dxeqn != null);
            yDefined = (dyeqn != null);
            zDefined = (dzeqn != null);
            wDefined = (dweqn != null);
            numOfVars = 0;
            
            if (xDefined) {
                xe = new CompiledStatement(dxeqn);
                numOfVars++;
            } else
                xe = null;
            
            if (yDefined) {
                ye = new CompiledStatement(dyeqn);
                numOfVars++;
            } else
                ye = null;
            
            if (zDefined) {
                ze = new CompiledStatement(dzeqn);
                numOfVars++;
            } else
                ze = null;
            
            if (wDefined) {
                we = new CompiledStatement(dweqn);
                numOfVars++;
            } else
                we = null;
            
            holder = new float[numOfVars];
        }
        
        public DiffEqGenerator(String dxeqn, String dyeqn, String dzeqn) {
            this(dxeqn,dyeqn,dzeqn,null);
        }
        
        public DiffEqGenerator(String dxeqn, String dyeqn) {
            this(dxeqn,dyeqn,null,null);
        }
        
        public DiffEqGenerator(String dxeqn) {
            this(dxeqn,null,null,null);
        }
        
        /**
         * Determines if a new point vector is made at each iteration
         * @param copy specifies if the points should be copied
         */
        public void setCopyPoints(boolean copy) {
            if (copy)
                holder = null;
            else
                holder = new float[numOfVars];
            
        }
        
        /**
         * Determines if the points are to be copied every iteration
         * @return true if a new array is to be allocated at every iteration
         */
        public boolean isCopyPoints() {
            return (holder == null);
        }
        
        public float[] nextIteration(float lastTime, float deltat, PointList wholeList) {
            //TODO:implement
            float[] lastElem = wholeList.getLastPoint();
            Map<String,Float> varMap = new HashMap<String,Float>();
            varMap.put("x",lastElem[0]);
            varMap.put("y",lastElem[1]);
            varMap.put("z",lastElem[2]);
            varMap.put("w",lastElem[3]);
            varMap.put("t",lastTime);
            
            float[]res = holder;
            if (res == null)
                res = new float[numOfVars];
            try {
                switch (numOfVars) {
                    case 4:
                        res[3] = lastElem[3] + this.we.evaluate(varMap)*deltat;
                    case 3:
                        res[2] =lastElem[2] + this.ze.evaluate(varMap)*deltat;
                    case 2:
                        res[1] = lastElem[1] + this.ye.evaluate(varMap)*deltat;
                    case 1:
                        res[0] = lastElem[0] + this.xe.evaluate(varMap)*deltat;
                        break;
                    default:
                        throw new SDDMException("Error iterating over differential equation - inappropriate number of variables");
                }
            } catch (EquationException e) {
                throw new SDDMException("Problem with differential equation evaluation: "+e.getMessage(),e);
            }
            System.err.println("Iter result: "+res[0]+", "+res[1]+", "+res[2]);
            return res;
        }
    }
    
    static class CompiledStatement {
        private String statement;
        
        enum BinaryOperation{ADD,SUB,MULT,DIV,MOD,POW};
        enum UnaryOperation{SIN,COS,TAN,EXP,LN,LOG10,ABS};
        enum StatementType{UPPER,LOWER,UNOP,VAR,CONST};
        
        private StatementType type = null;
        
        private String variableName = null;
        private Float value = null;
        private ArrayList<BinaryOperation> ops = null;
        private UnaryOperation unop = null;
        private ArrayList<CompiledStatement> innards = null;
        
        
        CompiledStatement(String statementIn) {
            this.statement = statementIn.trim();
            if (statement.equals("")) {
                type = StatementType.CONST;
                value = 0.0f;
            } else   {
                //Remove outer parenthesis if they are present
                if(statement.charAt(0)=='('&&statement.charAt(statement.length()-1)==')') {
                    statement = statement.substring(1,statement.length()-1);
                }
                
                //large-scale strategy:
                //1. split off parentheticals and compile
                //2. look for top-level binops, and split into expressions if they're found
                //3. look for second-level binops, and split them off
                //4. look for a unop, and compile innards if present
                //5. otherwise, set the constant value and mark as a const unless parsing fails
                //6. otherwise, set to variable with a name given in the string
                
                
                //look for upper-level binops
                for(int i = 0; i < statement.length();++i) {
                    i = skipParentheticals(this.statement,i);
                    
                    if (findUpperTierBinOp(statement.charAt(i))!=null) {
                        type = StatementType.UPPER;
                        ops = new ArrayList<BinaryOperation>();
                        innards = new ArrayList<CompiledStatement>();
                        break;
                    }
                }
                
                if (type == StatementType.UPPER) {
                    //TODO:test
                    int lastOpIndex = -1;
                    for (int i = 0; i < statement.length();++i) {
                        i = skipParentheticals(this.statement,i);
                        
                        BinaryOperation binOp = findUpperTierBinOp(statement.charAt(i));
                        if (binOp!= null) {
                            ops.add(binOp);
                            innards.add(new CompiledStatement(statement.substring(lastOpIndex+1,i)));
                            lastOpIndex = i;
                        }
                    }
                    innards.add(new CompiledStatement(statement.substring(lastOpIndex+1,statement.length())));
                } else {
                    //look for lower-level binops
                    for(int i = 0; i < statement.length();++i) {
                        i = skipParentheticals(this.statement,i);
                        
                        char curr = statement.charAt(i);
                        
                        if (curr == '(') {
                            //skip over parentheticals
                            int depth = 1;
                            while(depth > 0) {
                                curr = statement.charAt(++i);
                                if(curr == '(')
                                    ++depth;
                                if(curr == ')')
                                    --depth;
                            }
                        }
                        
                        else if (findLowerTierBinOp(curr)!=null) {
                            type = StatementType.LOWER;
                            ops = new ArrayList<BinaryOperation>();
                            innards = new ArrayList<CompiledStatement>();
                            break;
                        }
                    }
                    
                    if (type == StatementType.LOWER) {
                        //TODO:test
                        int lastOpIndex = -1;
                        for (int i = 0; i < statement.length();++i) {
                            i = skipParentheticals(this.statement,i);
                            
                            BinaryOperation binOp = findLowerTierBinOp(statement.charAt(i));
                            if (binOp!= null) {
                                ops.add(binOp);
                                innards.add(new CompiledStatement(statement.substring(lastOpIndex+1,i)));
                                lastOpIndex = i;
                            }
                        }
                        innards.add(new CompiledStatement(statement.substring(lastOpIndex+1,statement.length())));
                    } else {
                        StringBuffer sb = new StringBuffer(statement.trim());
                        unop = findUnOp(sb); //this removes the unop from the sb
                        if(unop != null) {
                            type = StatementType.UNOP;
                            innards = new ArrayList<CompiledStatement>();
                            
                            innards.add(new CompiledStatement(sb.toString()));
                            
                        } else {
                            try {
                                //try to get the constant number
                                float res = Float.parseFloat(statement);
                                type = StatementType.CONST;
                                value = new Float(res);
                            } catch (NumberFormatException e) {
                                //Must be a variable, as all other options have been eliminated
                                variableName = statement.trim();
                                type = StatementType.VAR;
                            }
                        }
                    }
                }
            }
        }
        
        private static int skipParentheticals(String stringToSkip, int startingIndex) {
            char curr = stringToSkip.charAt(startingIndex);
            
            if(curr!='(')
                return startingIndex;
            else {
                int i = startingIndex;
                int depth = 1;
                while(depth > 0) {
                    curr = stringToSkip.charAt(++i);
                    if(curr == '(')
                        ++depth;
                    if(curr == ')')
                        --depth;
                }
                return i;
            }
        }

        private static BinaryOperation findLowerTierBinOp(char charToTest) {
            BinaryOperation res = null;
            switch (charToTest) {
                case '*':
                    res = BinaryOperation.MULT;
                    break;
                case '/':
                    res = BinaryOperation.DIV;
                    break;
                case '%':
                    res = BinaryOperation.MOD;
                    break;
                case '^':
                    res = BinaryOperation.POW;
                    break;
            }
            return res;
        }
        
        private static BinaryOperation findUpperTierBinOp(char charToTest) {
            BinaryOperation res = null;
            switch (charToTest) {
                case '+':
                    res = BinaryOperation.ADD;
                    break;
                case '-':
                    res = BinaryOperation.SUB;
                    break;
            }
            return res;
        }
        
        private static UnaryOperation findUnOp(StringBuffer sbToTest) {
            //input is trimmed
            //should alter the input string to remove the unop - only the internal expression should be left
            //UnaryOperation{SIN,COS,TAN,EXP,LN,LOG10,ABS};
            final String[] matchStr = {"sin","cos","tan","exp","ln","loge","log10","log","abs"};
            final UnaryOperation[] matchUnOp = {UnaryOperation.SIN,UnaryOperation.COS,UnaryOperation.TAN,UnaryOperation.EXP,UnaryOperation.LN,UnaryOperation.LN,UnaryOperation.LOG10,UnaryOperation.SIN.LOG10,UnaryOperation.ABS};
            //TODO:test
            for (int i = 0; i < matchStr.length;++i) {
                if(matchStr[i].length() <= sbToTest.length() && sbToTest.substring(0,matchStr[i].length()).equalsIgnoreCase(matchStr[i])) {
                    sbToTest.delete(0,matchStr[i].length()+1);
                    sbToTest.deleteCharAt(sbToTest.length()-1);
                    return matchUnOp[i];
                }
                
            }
            return null;
        }
        
        int getNumOfVars() {
            if(this.type == StatementType.CONST)
                return 0;
            else if (this.type == StatementType.VAR)
                return 1;
            else
                return getVars().size();
        }
        
        Set<String> getVars() {
            TreeSet<String> vars = new TreeSet<String>();
            switch(this.type) {
                case UPPER:
                case LOWER:
                case UNOP:
                    for (CompiledStatement cs : this.innards)
                        vars.addAll(cs.getVars());
                    break;
                case VAR:
                    vars.add(this.variableName);
                    break;
            }
            return vars;
        }
        
        float evaluate(Map<String,Float> varMap) throws EquationException {
            float[] varsIn = new float[this.getNumOfVars()];
            int i = 0;
            
            for(String s : this.getVars()) {
                Float f = varMap.get(s);
                if(f == null)
                    throw new EquationException("Did not provide value for variable "+s);
                else
                    varsIn[i++] = f;
            }
            
            return evaluate(varsIn);
        }
        
        float evaluate(float[] varsIn) throws EquationException {
            if (varsIn.length != getNumOfVars())
                throw new EquationException("Number of vars provided:"+varsIn.length+" not equal to number of variables necessary for this equation:"+getNumOfVars());
            
            switch(type) {
                case LOWER:
                case UPPER:
                    ArrayList<float[]> varsToEval= new ArrayList<float[]>();
                    for(CompiledStatement cs : innards) {
                        //TODO: be smarter - need to match by name
                        String[] varNames = new String[this.getNumOfVars()];
                        int varNum = 0;
                        for (String s : getVars())
                            varNames[varNum++] = s;
                        
                        float[] newVars = new float[cs.getNumOfVars()];
                        int i = 0;
                        for (String s : cs.getVars()) {
                            for (int j = 0; j < varNames.length;++j) {
                                if (varNames[j] == s)
                                    newVars[i++] = varsIn[j];
                            }
                        }
                        varsToEval.add(newVars);
                    }
                    //TODO: implement checking so all vars are accounted for
                    
                    //do first two pairs
                    float result = applyBinOp(ops.get(0),innards.get(0).evaluate(varsToEval.get(0)),innards.get(1).evaluate(varsToEval.get(1)));
                    //do all the rest
                    for(int i = 1; i < ops.size();++i) {
                        BinaryOperation bo = ops.get(i);
                        result = applyBinOp(bo,result,innards.get(i+1).evaluate(varsToEval.get(i+1)));
                    }
                    
                    return result;
                case UNOP:
                    return applyUnOp(this.unop,this.innards.get(0).evaluate(varsIn));
                case VAR:
                    return varsIn[0];
                case CONST:
                    return this.value;
                    
            }
            return 0;
        }
        
        private float applyBinOp(BinaryOperation bo, float leftValue, float rightValue) {
            float result = 0;
            switch (bo) {
                case ADD:
                    result = leftValue + rightValue;
                    break;
                case SUB:
                    result = leftValue - rightValue;
                    break;
                case MULT:
                    result = leftValue * rightValue;
                    break;
                case DIV:
                    result = leftValue / rightValue;
                    break;
                case MOD:
                    result = leftValue % rightValue;
                    break;
                case POW:
                    result = (float)Math.pow(leftValue,rightValue);
                    break;
                default:
                    throw new SDDMException("Unrecognized binary operation type in compiled equation:"+bo.name());
            }
            return result;
        }
        
        private float applyUnOp(UnaryOperation uo, float value) {
            float result = 0;
            switch (uo) {
                case SIN:
                    result = (float) Math.sin(value);
                    break;
                case COS:
                    result = (float) Math.cos(value);
                    break;
                case TAN:
                    result = (float) Math.tan(value);
                    break;
                case EXP:
                    result = (float) Math.exp(value);
                    break;
                case LN:
                    result = (float) Math.log(value);
                    break;
                case LOG10:
                    result = (float) Math.log10(value);
                    break;
                case ABS:
                    result = Math.abs(value);
                    break;
                default:
                    throw new SDDMException("Unrecognized unary operation type in compiled equation:"+uo.name());
            }
            return result;
        }
        
        public String toString() {
            return this.statement;
        }
        
    }
    
    static class EquationException extends Exception {
        EquationException() {
            super();
        }
        EquationException(String s) {
            super(s);
        }
    }
    
}
