/*
 * ParametricRosslerTypePanel.java
 *
 * Created on August 3, 2006, 11:40 AM
 */

package sddm.modules;

import sddm.*;
import sddm.data.*;

/**
 *
 * @author  Erik J. Tollerud
 */
public class ParametricRosslerTypePanel extends ParametricModuleSetupPanel.TypePanel {
    ParametricModuleSetupPanel parent;
    
    /**
     * Creates new form ParametricRosslerTypePanel
     */
    public ParametricRosslerTypePanel(ParametricModuleSetupPanel sp) {
        super();
        this.parent = sp;
        initComponents();
    }
    
    public ParametricRosslerTypePanel(ParametricModuleSetupPanel sp, ParametricSystem initialSystem) {
        this(sp);
        if (initialSystem != null) {
            if (initialSystem.getGen() instanceof UniformParametricGenerator) {
                UniformParametricGenerator upg = (UniformParametricGenerator)initialSystem.getGen();
                UniformParametricGenerator.Generator gen = upg.getGen();
                if (gen instanceof RosslerGenerator) {
                    RosslerGenerator rg = (RosslerGenerator)gen;
                    this.aField.setValue(rg.a);
                    this.bField.setValue(rg.b);
                    this.cField.setValue(rg.c);
                    
                    this.deltaTField.setValue(upg.getDeltat());
                    this.x0Field.setValue(upg.getX0());
                    this.y0Field.setValue(upg.getY0());
                    this.z0Field.setValue(upg.getZ0());
                    
                    this.itersField.setValue(initialSystem.getCurrentIteration());
                }
            }
        }
    }
    
    public sddm.data.ParametricSystem makeSystem() {
        float a = ((Number)this.aField.getValue()).floatValue();
        float b = ((Number)this.bField.getValue()).floatValue();
        float c = ((Number)this.cField.getValue()).floatValue();
        
        float dt = ((Number)this.deltaTField.getValue()).floatValue();
        float x0 = ((Number)this.x0Field.getValue()).floatValue();
        float y0 = ((Number)this.y0Field.getValue()).floatValue();
        float z0 = ((Number)this.z0Field.getValue()).floatValue();
        
        UniformParametricGenerator upg = new UniformParametricGenerator(dt,new RosslerGenerator(a,b,c),x0,y0,z0,0.0f);
        
        ParametricSystem ps = new ParametricSystem(upg);
        
        int iters = ((Number)this.itersField.getValue()).intValue();
        ps.iterate(iters);
        
        return ps;
    }
    
    /** This method is called from within the constructor to
     * initialize the form.
     * WARNING: Do NOT modify this code. The content of this method is
     * always regenerated by the Form Editor.
     */
    // <editor-fold defaultstate="collapsed" desc=" Generated Code ">//GEN-BEGIN:initComponents
    private void initComponents() {
        aField = new javax.swing.JFormattedTextField(java.text.NumberFormat.getInstance());
        bField = new javax.swing.JFormattedTextField(java.text.NumberFormat.getInstance());
        cField = new javax.swing.JFormattedTextField(java.text.NumberFormat.getInstance());
        jLabel1 = new javax.swing.JLabel();
        x0Field = new javax.swing.JFormattedTextField(java.text.NumberFormat.getInstance());
        y0Field = new javax.swing.JFormattedTextField(java.text.NumberFormat.getInstance());
        z0Field = new javax.swing.JFormattedTextField(java.text.NumberFormat.getInstance());
        jLabel2 = new javax.swing.JLabel();
        jLabel3 = new javax.swing.JLabel();
        jLabel4 = new javax.swing.JLabel();
        itersField = new javax.swing.JFormattedTextField(java.text.NumberFormat.getInstance());
        deltaTField = new javax.swing.JFormattedTextField(java.text.NumberFormat.getInstance());
        jLabel5 = new javax.swing.JLabel();
        jLabel6 = new javax.swing.JLabel();

        aField.setText("num");
        aField.setValue(0.2f);

        bField.setText("num");
        bField.setValue(0.2f);

        cField.setText("num");
        cField.setValue(5.7f);

        jLabel1.setText("Initial Point");

        x0Field.setText("num");
        x0Field.setValue(1.0f);

        y0Field.setText("num");
        y0Field.setValue(1.0f);

        z0Field.setText("num");
        z0Field.setValue(1.0f);

        jLabel2.setText("a");

        jLabel3.setText("b");

        jLabel4.setText("c");

        itersField.setText("num");
        itersField.setValue(1000);

        deltaTField.setText("num");
        deltaTField.setValue(0.01f);

        jLabel5.setText("deltat");

        jLabel6.setText("initial iterations");

        org.jdesktop.layout.GroupLayout layout = new org.jdesktop.layout.GroupLayout(this);
        this.setLayout(layout);
        layout.setHorizontalGroup(
            layout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
            .add(layout.createSequentialGroup()
                .addContainerGap()
                .add(layout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
                    .add(layout.createSequentialGroup()
                        .add(jLabel1)
                        .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                        .add(x0Field, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE)
                        .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                        .add(y0Field, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE)
                        .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                        .add(z0Field, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE))
                    .add(layout.createSequentialGroup()
                        .add(jLabel2)
                        .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                        .add(aField, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE)
                        .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                        .add(jLabel3)
                        .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                        .add(bField, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE)
                        .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                        .add(jLabel4)
                        .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                        .add(cField, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE))
                    .add(layout.createSequentialGroup()
                        .add(jLabel5)
                        .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                        .add(deltaTField, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE)
                        .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                        .add(jLabel6)
                        .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                        .add(itersField, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE)))
                .addContainerGap(org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE))
        );
        layout.setVerticalGroup(
            layout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
            .add(layout.createSequentialGroup()
                .addContainerGap()
                .add(layout.createParallelGroup(org.jdesktop.layout.GroupLayout.BASELINE)
                    .add(jLabel1)
                    .add(x0Field, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE)
                    .add(y0Field, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE)
                    .add(z0Field, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE))
                .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                .add(layout.createParallelGroup(org.jdesktop.layout.GroupLayout.BASELINE)
                    .add(jLabel2)
                    .add(aField, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE)
                    .add(jLabel3)
                    .add(bField, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE)
                    .add(jLabel4)
                    .add(cField, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE))
                .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
                .add(layout.createParallelGroup(org.jdesktop.layout.GroupLayout.BASELINE)
                    .add(jLabel5)
                    .add(deltaTField, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE)
                    .add(jLabel6)
                    .add(itersField, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE))
                .addContainerGap(org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE))
        );
    }// </editor-fold>//GEN-END:initComponents
    
    
    // Variables declaration - do not modify//GEN-BEGIN:variables
    private javax.swing.JFormattedTextField aField;
    private javax.swing.JFormattedTextField bField;
    private javax.swing.JFormattedTextField cField;
    private javax.swing.JFormattedTextField deltaTField;
    private javax.swing.JFormattedTextField itersField;
    private javax.swing.JLabel jLabel1;
    private javax.swing.JLabel jLabel2;
    private javax.swing.JLabel jLabel3;
    private javax.swing.JLabel jLabel4;
    private javax.swing.JLabel jLabel5;
    private javax.swing.JLabel jLabel6;
    private javax.swing.JFormattedTextField x0Field;
    private javax.swing.JFormattedTextField y0Field;
    private javax.swing.JFormattedTextField z0Field;
    // End of variables declaration//GEN-END:variables
    
    public static class RosslerGenerator implements sddm.data.UniformParametricGenerator.Generator {
        private float a;
        private float b;
        private float c;
        
        private float[] holder;
        
        public RosslerGenerator(float a, float b, float c) {
            this.a = a;
            this.b = b;
            this.c = c;
            holder = new float[4];
        }
        public float[] nextIteration(float lastTime, float deltat, PointList wholeList) {
            //Rossler map:
            //dx/dt = -y-z
            //dy/dt = x+ay
            //dz/dt = b+z(x-c)
            
            float[] res = holder;
            float[] lastElem = wholeList.getLastPoint();
            
            res[3] = lastElem[3];
            
            res[0] = lastElem[0] - (lastElem[1] + lastElem[2])*deltat;
            res[1] = lastElem[1] + (lastElem[0]+ this.a*lastElem[1])*deltat;
            res[2] = lastElem[2] + (this.b+lastElem[2]*(lastElem[0] - this.c))*deltat;
            
            return res;
        }
        public String toString() {
            return "Lorenz-Salzman equations with a="+this.a+", b="+this.b+", c="+this.c;
        }
    }
}
