/*
 * StereoImageModule.java
 *
 * Created on July 15, 2006, 2:03 AM
 *
 * sample command line: -laf native -sm a -mod si -lil "C:\Documents and Settings\Erik\Desktop\stereo pairs\leftlist.txt" -ril "C:\Documents and Settings\Erik\Desktop\stereo pairs\rightlist.txt"
 */

package sddm.modules;

import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import javax.media.opengl.*;
import com.sun.opengl.util.texture.*;
import java.io.*;
import java.util.*;
import java.awt.image.*;
import javax.swing.JPanel;
import java.awt.image.BufferedImage;
import sddm.*;
import sddm.SDDModule;


/**
 * TODO: add image adders/removers, instead of all the work being done in constructors
 * @author Erik J. Tollerud
 */
public class StereoImageModule  extends SDDModule {
    BufferedImage[] leftImages;
    BufferedImage[] rightImages;
    Texture leftTexture, rightTexture;
    private boolean advanceOnDisplay = false;
    private boolean advancel = true;
    private boolean advancer = true;
    private boolean redrawCurrent = false;
    private boolean reshapeToActualImageRatio = false;
    private int imageIndex;
    //TODO: implement as remembered array
    //private float[] leftRot;
    //private float[] rightRot;
    //private float[] leftXTrans;
    //private float[] rightXTrans;
    //private float[] leftYTrans;
    //private float[] rightYTrans;
    private float rotR = 0;
    private float transXR = 0;
    private float transYR = 0;
    private float scaleXR = 1;
    private float scaleYR = 1;
    private float rotL = 0;
    private float transXL = 0;
    private float transYL = 0;
    private float scaleXL = 1;
    private float scaleYL = 1;
    
    public enum MouseInteraction{ROT,TRANS,SCALE};
    private MouseInteraction mouseInteraction = MouseInteraction.ROT;
    private float mouseRotStep = 1;
    private float mouseTransStep = 0.02f;
    private float mouseScaleStep = 0.02f;
    
    /** Creates a new instance of StereoImageModule */
    public StereoImageModule() {
        super();
        
        leftImages = new BufferedImage[1];
        rightImages = new BufferedImage[1];
        imageIndex = 0;
        
        //left pixel is set to blue
        leftImages[0] = new BufferedImage(1,1,BufferedImage.TYPE_INT_ARGB);
        leftImages[0].setRGB(0,0,16711680);
        
        //right pixel is set to red
        rightImages[0] = new BufferedImage(1,1,BufferedImage.TYPE_INT_ARGB);
        rightImages[0].setRGB(0,0,255);
    }
    public StereoImageModule(StereoCamera camera) {
        super(camera);
        
        leftImages = new BufferedImage[1];
        rightImages = new BufferedImage[1];
        imageIndex = 0;
        
        //left pixel is set to blue
        leftImages[0] = new BufferedImage(1,1,BufferedImage.TYPE_INT_ARGB);
        leftImages[0].setRGB(0,0,16711680);
        
        //right pixel is set to red
        rightImages[0] = new BufferedImage(1,1,BufferedImage.TYPE_INT_ARGB);
        rightImages[0].setRGB(0,0,255);
        
    }
    
    public StereoImageModule(BufferedImage leftImage, BufferedImage rightImage) {
        super();
        
        leftImages = new BufferedImage[1];
        rightImages = new BufferedImage[1];
        imageIndex = 0;
        
        leftImages[0] = leftImage;
        rightImages[0] = rightImage;
        
    }
    public StereoImageModule(BufferedImage[] leftImages, BufferedImage[] rightImages) throws SDDMException{
        super();
        if(leftImages.length != rightImages.length)
            throw new SDDMException("Stereo image sets aren't the same length for both eyes");
        imageIndex = 0;
        
        this.leftImages = leftImages;
        this.rightImages = rightImages;
        
    }
    
    public StereoImageModule(File leftImage, File rightImage) throws IOException {
        super();
        
        leftImages = new BufferedImage[1];
        rightImages = new BufferedImage[1];
        imageIndex = 0;
        
        leftImages[0] = javax.imageio.ImageIO.read(leftImage);
        rightImages[0] = javax.imageio.ImageIO.read(rightImage);
        
    }
    
    public StereoImageModule(File[] leftImages, File[] rightImages) throws IOException, SDDMException {
        super();
        if(leftImages.length != rightImages.length)
            throw new SDDMException("Stereo image sets aren't the same length for both eyes");
        
        this.leftImages = new BufferedImage[leftImages.length];
        this.rightImages = new BufferedImage[rightImages.length];
        imageIndex = 0;
        
        for (int i = 0; i < leftImages.length; ++i) {
            this.leftImages[i] = javax.imageio.ImageIO.read(leftImages[i]);
            this.rightImages[i] = javax.imageio.ImageIO.read(rightImages[i]);
        }
        
        
        
    }
    
    public void init(GL gl) {
        this.setDisplayListOn(false);
        gl.glTexEnvf(gl.GL_TEXTURE_ENV, gl.GL_TEXTURE_ENV_MODE,gl.GL_DECAL);
        advancel = true;
        advancer = true;
    }
    
    public void preDraw(GL gl) {
        gl.glMatrixMode(GL.GL_PROJECTION);
        gl.glLoadIdentity();
        gl.glMatrixMode(GL.GL_MODELVIEW);
        gl.glLoadIdentity();
        if(this.cam.isRight()) {
            gl.glTranslatef(this.transXR,this.transYR,0);
            gl.glRotatef(this.rotR,0,0,-1);
            gl.glScalef(this.scaleXR,this.scaleYR,0);
        }
        else {
            gl.glTranslatef(this.transXL,this.transYL,0);
            gl.glRotatef(this.rotL,0,0,-1);
            gl.glScalef(this.scaleXL,this.scaleYL,0);
        }
    }
    
    public void drawGeometry(GL gl) {
        boolean onRight = this.cam.isRight();
        
        //TextureIO method
        Texture mainTexture;
        if (onRight) {
            if (advancer) {
                if (rightTexture != null)
                    rightTexture.dispose();
                rightTexture = TextureIO.newTexture(this.rightImages[imageIndex],false);
                advancer = false;
            }
            mainTexture = rightTexture;
        } else {
            if (advancel) {
                if (leftTexture != null)
                    leftTexture.dispose();
                leftTexture = TextureIO.newTexture(this.leftImages[imageIndex],false);
                advancel = false;
            }
            mainTexture = leftTexture;
        }

        mainTexture.enable();
        mainTexture.bind();
        
        
        TextureCoords tc = mainTexture.getImageTexCoords();
        
        if (reshapeToActualImageRatio) {
            BufferedImage currImg;
            if(onRight) 
                currImg = this.rightImages[imageIndex];
            else
                currImg = this.leftImages[imageIndex];
            
            float maxRight = 1;
            float maxBottom = 1;
            float imgRatio = (float)currImg.getWidth()/(float)currImg.getHeight();
            
            int[] vp = new int[4];
            gl.glGetIntegerv(GL.GL_VIEWPORT,vp,0);
            float viewPortRatio = (float)vp[2]/(float)vp[3]; //TODO:fix
            System.err.println("imgr:"+imgRatio+"vpr:"+viewPortRatio);
            if (imgRatio > viewPortRatio)
                maxBottom = viewPortRatio/imgRatio;
            else
                maxRight = imgRatio/viewPortRatio;
            
            gl.glBegin(GL.GL_QUADS);
            //lower-left
            gl.glTexCoord2f(tc.left(),tc.bottom());
            gl.glVertex2f(-1,-maxBottom);
            //lower-right
            gl.glTexCoord2f(tc.right(),tc.bottom());
            gl.glVertex2f(maxRight,-maxBottom);
            //upper-right
            gl.glTexCoord2f(tc.right(),tc.top());
            gl.glVertex2f(maxRight,1);
            //upper-left
            gl.glTexCoord2f(tc.left(),tc.top());
            gl.glVertex2f(-1,1);
            gl.glEnd();
        } else {
            gl.glBegin(GL.GL_QUADS);
            //lower-left
            gl.glTexCoord2f(tc.left(),tc.bottom());
            gl.glVertex2f(-1,-1);
            //lower-right
            gl.glTexCoord2f(tc.right(),tc.bottom());
            gl.glVertex2f(1,-1);
            //upper-right
            gl.glTexCoord2f(tc.right(),tc.top());
            gl.glVertex2f(1,1);
            //upper-left
            gl.glTexCoord2f(tc.left(),tc.top());
            gl.glVertex2f(-1,1);
            gl.glEnd();
        }
        
        mainTexture.disable();
         
        
        /*//self-made texture
        BufferedImage bi;
        if (onRight)
            bi = this.rightImages[imageIndex];
        else
            bi = this.leftImages[imageIndex];
        
        
        //int[] arr = new int[bi.getWidth()*bi.getHeight()];
        //int count = 0;
        //for(int i = 0; i < bi.getWidth();i++) {
        //    for(int j = 0;  j < bi.getHeight();j++) {
        //        arr[count++] = bi.getRGB(i,j);  
        //    }
        //}
        //    
        //java.nio.Buffer buffer = java.nio.IntBuffer.wrap(arr);
        java.nio.Buffer buffer = java.nio.IntBuffer.wrap(bi.getRGB(0,0,bi.getWidth(),bi.getHeight(),null,0,bi.getWidth()));
        //DataBuffer db = bi.getRaster().getDataBuffer();
        //java.nio.Buffer buffer = null;
        //int type = 0;
       //switch(db.getDataType()) {
       //    case DataBuffer.TYPE_BYTE:
       //        type = GL.GL_UNSIGNED_BYTE;
       //        buffer = java.nio.ByteBuffer.wrap(((java.awt.image.DataBufferByte)db)..getData());
       //        break;
       //    case DataBuffer.TYPE_DOUBLE:
       //        type = GL.GL_DOUBLE;
       //        buffer = java.nio.DoubleBuffer.wrap(((java.awt.image.DataBufferDouble)db).getData());
       //        break;
       //    case DataBuffer.TYPE_FLOAT:
       //        type = GL.GL_FLOAT;
       //        buffer = java.nio.FloatBuffer.wrap(((java.awt.image.DataBufferFloat)db).getData());
       //        break;
       //    case DataBuffer.TYPE_INT:
       //        type = GL.GL_INT;
       //        buffer = java.nio.IntBuffer.wrap(((java.awt.image.DataBufferInt)db).getData());
       //        break;
       //    case DataBuffer.TYPE_SHORT:
       //        type = GL.GL_SHORT;
       //        buffer = java.nio.ShortBuffer.wrap(((java.awt.image.DataBufferShort)db).getData());
       //        break;
       //    case DataBuffer.TYPE_USHORT:
       //         type = GL.GL_UNSIGNED_SHORT;
       //        buffer = java.nio.ShortBuffer.wrap(((java.awt.image.DataBufferUShort)db).getData());
          //     
       //}
        
        
        
        gl.glTexParameterf(GL.GL_TEXTURE_2D,GL.GL_TEXTURE_MAG_FILTER,GL.GL_NEAREST);
        gl.glTexParameterf(GL.GL_TEXTURE_2D,GL.GL_TEXTURE_MIN_FILTER,GL.GL_NEAREST);
        
        gl.glTexEnvf(gl.GL_TEXTURE_ENV, gl.GL_TEXTURE_ENV_MODE,gl.GL_DECAL);
        
        gl.glTexImage2D(GL.GL_TEXTURE_2D,0,GL.GL_RGBA,bi.getWidth(),bi.getHeight(),0,GL.GL_ABGR_EXT,GL.GL_UNSIGNED_INT_8_8_8_8,buffer);
        //gl.glTexImage2D(GL.GL_TEXTURE_2D,0,GL.GL_RGBA,bi.getWidth(),bi.getHeight(),0,GL.GL_RGB,type,buffer);
        
        //Have to flip vertically for images to be correct ?? Also, should be ARGB instead of ABGR
        gl.glBegin(GL.GL_QUADS);
        //lower-left
        gl.glTexCoord2f(0,1);
        gl.glVertex2f(-1,-1);
        //lower-right
        gl.glTexCoord2f(1,1);
        gl.glVertex2f(1,-1);
        //upper-right
        gl.glTexCoord2f(1,0);
        gl.glVertex2f(1,1);
        //upper-left
        gl.glTexCoord2f(0,0);
        gl.glVertex2f(-1,1);
        gl.glEnd(); 
        
        gl.glDisable(GL.GL_TEXTURE_2D);*/
        
        //have to check which side so that it isn't advanced twice per dispaly
        if(this.advanceOnDisplay && !onRight)
            advance();
    }
    
    public static SDDModule makeModuleFromArgs(String modName, java.util.HashMap<String,Integer> switchMap, String[] args, StereoCamera camera) throws  ParsingException{
        //takes modName si, stereoimage, or stereoimages
        //TODO: add functionality for multiple images/folders
        //switches accepted:
        //-*leftimagefile|lif|li
        //-*rightimagefile|rif|ri
        //| (single file has precidence and overrides file list)
        //-leftimagefilelist|leftimagelist|lil|lfl
        //-rightimagefilelist|rightimagelist|ril|rfl
        //
        //-[advanceondisplay|aod] (true|t)|(false|f)
        
        if (modName.matches("si|stereoimage|stereoimages")) {
            boolean singleImages = true;
            Integer leftIndx = switchMap.remove("-leftimagefile");
            if (leftIndx == null)
                leftIndx = switchMap.remove("-lif");
            if (leftIndx == null)
                leftIndx = switchMap.remove("-li");
            
            Integer rightIndx = switchMap.remove("-rightimagefile");
            if (rightIndx == null)
                rightIndx = switchMap.remove("-rif");
            if (rightIndx == null)
                rightIndx = switchMap.remove("-ri");
            
            //check for file lists
            if (leftIndx == null || rightIndx == null) {
                singleImages = false;
                
                
                leftIndx = switchMap.remove("-leftimagefilelist");
                if (leftIndx == null)
                    leftIndx = switchMap.remove("-leftimagelist");
                if (leftIndx == null)
                    leftIndx = switchMap.remove("-lil");
                if (leftIndx == null)
                    leftIndx = switchMap.remove("-lfl");
                
                rightIndx = switchMap.remove("-rightimagefilelist");
                if (rightIndx == null)
                    rightIndx = switchMap.remove("-rightimagelist");
                if (rightIndx == null)
                    rightIndx = switchMap.remove("-ril");
                if (rightIndx == null)
                    rightIndx = switchMap.remove("-rfl");
                
                // if no files or file lists, throw exception
                if(leftIndx == null || rightIndx == null)
                    throw new ParsingException("no images specified for stereo image module");
            }
            
            File leftFile = new File(args[leftIndx+1]);
            File rightFile = new File(args[rightIndx+1]);
            StereoImageModule m = null;
            try {
                if(singleImages)
                    m = new StereoImageModule(leftFile,rightFile);
                else {
                    File leftwd = leftFile.getParentFile(); 
                    File rightwd = rightFile.getParentFile();
                    File[] leftFileList = new File[1];
                    File[] rightFileList = new File[1];
                    
                    ArrayList<File> leftAList= new ArrayList<File>();
                    ArrayList<File> rightAList = new ArrayList<File>();
                    
                    //TODO: add to list
                    LineNumberReader lread = new LineNumberReader(new FileReader(leftFile));
                    for(String s = lread.readLine();s!=null;s = lread.readLine()) 
                        leftAList.add(new File(s));
                    
                    LineNumberReader rread = new LineNumberReader(new FileReader(rightFile));
                    for(String s = rread.readLine();s!=null;s = rread.readLine()) 
                        rightAList.add(new File(s));
                    
                    if(leftAList.size() != rightAList.size())
                            throw new ParsingException("File Lists Supplied at command line for Stereo Image Module are not identical lengths");
                    
                    leftFileList = leftAList.toArray(leftFileList);
                    rightFileList = rightAList.toArray(rightFileList);
                    
                    m = new StereoImageModule(leftFileList,rightFileList);
                }
                
                Integer aodIndx = switchMap.remove("-aod");
                if (aodIndx == null)
                    aodIndx = switchMap.remove("-advanceondisplay");
                if (aodIndx != null) {
                    String aodStr = args[aodIndx+1];
                    if (aodStr.matches("t|true"))
                        m.setAdvanceOnDisplay(true);
                    if (aodStr.matches("f|false"))
                        m.setAdvanceOnDisplay(false);
                }
            } catch(IOException e) {
                throw new ParsingException("IOException loading stereo image modules files:"+e.getMessage(),e);
            }

            m.setCamera(camera);
            return m;
        }
        return null;
    }
    
    public float getMinFeatureSizeHint() {
        if (this.leftImages != null && this.rightImages != null) {
            BufferedImage left = this.leftImages[imageIndex];
            BufferedImage right = this.rightImages[imageIndex];
            
            float maxPixels = Math.max(Math.max(left.getHeight(),left.getWidth()),Math.max(right.getHeight(),right.getWidth()));
            
            return 2.0f/maxPixels;
        }
        return super.getMinFeatureSizeHint();
    }
    
    public static String moduleTitle() {
        return "Stereo Image Pair Module";
    }
    
    public StereoImageModuleControlPanel newModuleControlPanel() {
        return new StereoImageModuleControlPanel(this,cam);
    }
    
    
    public static StereoImageModuleSetupPanel newStartupPanel() {
        return new StereoImageModuleSetupPanel();
    }
    
    public void setAdvanceOnDisplay(boolean advanceOnDisplay) {
        this.advanceOnDisplay = advanceOnDisplay;
    }
    
    public boolean isAdvanceOnDisplay() {
        return advanceOnDisplay;
    }
    
    public int getNumOfImagePairs() {
        return leftImages.length;
    }
    
    public void advance() {
        if(++imageIndex >= leftImages.length)
            imageIndex = 0;
        advancer = true;
        advancel = true;
    }
    
    public void rotate(float degreesToRotate, boolean right) {
        if(right)
            this.rotR+=degreesToRotate;
        else
            this.rotL+=degreesToRotate;
    }
    
    public float getRotation(boolean right) {
        if (right)
            return this.rotR;
        else 
            return this.rotL;
    }
    
    public void translateX(float screenFractionToTranslate, boolean right) {
        if (right)
            this.transXR+=screenFractionToTranslate;
        else
            this.transXL+=screenFractionToTranslate;
    }
    
    public float getXTranslation(boolean right) {
        if (right)
            return this.transXR;
        else
            return this.transXL;
    }
    
    public void translateY(float screenFractionToTranslate, boolean right) {
        if (right)
            this.transYR+=screenFractionToTranslate;
        else
            this.transYL+=screenFractionToTranslate;
    }
    
    public float getYTranslation(boolean right) {
        if (right)
            return this.transYR;
        else
            return this.transYL;
    }
    
    public void scaleX(float screenFractionToScale, boolean right) {
        if (right)
            this.scaleXR+=screenFractionToScale;
        else
            this.scaleXL+=screenFractionToScale;
    }
    
    public float getXScale(boolean right) {
        if (right)
            return this.scaleXR;
        else
            return this.scaleXL;
    }
    
    public void scaleY(float screenFractionToScale, boolean right) {
        if (right)
            this.scaleYR+=screenFractionToScale;
        else
            this.scaleYL+=screenFractionToScale;
    }
    
    public float getYScale(boolean right) {
        if (right)
            return this.scaleYR;
        else
            return this.scaleYL;
    }

    public void setReshapeToActualImageRatio(boolean reshapeToActualImageRatio) {
        this.reshapeToActualImageRatio = reshapeToActualImageRatio;
    }
    
    public boolean isReshapeToActualImageRatio() {
        return reshapeToActualImageRatio;
    }

    public void setMouseRotStep(float mouseRotStep) {
        this.mouseRotStep = mouseRotStep;
    }

    public float getMouseRotStep() {
        return mouseRotStep;
    }

    public void setMouseScaleStep(float mouseScaleStep) {
        this.mouseScaleStep = mouseScaleStep;
    }

    public float getMouseScaleStep() {
        return mouseScaleStep;
    }

    public void setMouseTransStep(float mouseTransStep) {
        this.mouseTransStep = mouseTransStep;
    }

    public float getMouseTransStep() {
        return mouseTransStep;
    }

    public StereoImageModule.MouseInteraction getMouseInteraction() {
        return mouseInteraction;
    }
    
    public void setMouseInteraction(StereoImageModule.MouseInteraction mouseInteraction) {
        this.mouseInteraction = mouseInteraction;
    }


    

    
    public int getImageIndex() {
        //TODO: test
        if(this.advanceOnDisplay)
            return imageIndex - 1;
        return imageIndex;
    }
   
    int oldMouseX;
    int oldMouseY;
    int mouseMoving;
    
    public void attachEventListeners(StereoCameraControlFrame frame) {
        super.attachEventListeners(frame);
        
        final StereoCameraControlFrame frameF = frame;
        frame.addKeyListenerToJFrames(new java.awt.event.KeyAdapter() {
            public void keyPressed(KeyEvent e) {
                switch(e.getKeyChar()) {
                    case 'a':
                        advance();
                        frameF.redisplay();
                        break;
                    case 'd':
                        float factor = 0.5f;
                        String factorStr = javax.swing.JOptionPane.showInputDialog("What factor should be used for computing differential filter? (1=all filter, 0=no change)?","0.5");
                        if (factorStr != null) {
                            factor = Float.parseFloat(factorStr);
                        }
                        diffFilter(leftImages[imageIndex],factor);
                        diffFilter(rightImages[imageIndex],factor);
                        recompileDisplayList();
                        advancel = true;
                        advancer = true;
                        frameF.redisplay();
                        break;
                }
            }
        });
        
        //remove all mouse events - they are no longer applicable
        for (java.awt.event.MouseListener ml : frame.getMouseListeners())
            frame.removeMouseListener(ml);
        for (java.awt.event.MouseMotionListener mml : frame.getMouseMotionListeners())
            frame.removeMouseMotionListener(mml);
        for (java.awt.event.MouseWheelListener mwl : frame.getMouseWheelListeners())
            frame.removeMouseWheelListener(mwl);
        
        frame.addMouseListenerToJFrames(new java.awt.event.MouseAdapter() {
            public void mouseReleased(MouseEvent e) {
                mouseMoving = -1;
            }

            public void mousePressed(MouseEvent e) {
                oldMouseX = e.getX();
                oldMouseY = e.getY();
                mouseMoving = e.getButton();
            }
            
        });
        
        frame.addMouseMotionListenerToJFrames(new java.awt.event.MouseMotionAdapter() {
            public void mouseMoved(MouseEvent e) {
                super.mouseMoved(e);
            }

            public void mouseDragged(MouseEvent e) {
                int deltaX = e.getX() - oldMouseX;
                int deltaY = e.getY() - oldMouseY;
                boolean b1 = mouseMoving != e.BUTTON1;
                
                switch(mouseInteraction) {
                    case ROT:
                        rotate((deltaX+deltaY)*mouseRotStep, b1);
                        break;
                    case TRANS:
                        translateX(deltaX*mouseTransStep, b1);
                        translateY(-deltaY*mouseTransStep, b1);
                        break;
                    case SCALE:
                        scaleX(deltaX*mouseScaleStep, b1);
                        scaleY(-deltaY*mouseScaleStep, b1);
                        break;
                }
                
                oldMouseX = e.getX();
                oldMouseY = e.getY();
                frameF.redisplay();
            }
            
        });
        
    }
    
    //------------filters----------
    
    /**
     * Applies differential filter - each pixel on a line is set from difference between it and the previous 
     * @param img the image to apply the filter to.
     * @param factor the amount to use of the differential (i.e. 1 is all derivative, 0 leaves file unchanged).  If negative, goes along vertical scanlines.
     */
    private void diffFilter(BufferedImage img, float factor) {
        boolean horiz;
        if (factor >0) {
            //horizontal lines
            horiz = true;
            
        }
        else if (factor <0) {
            //vertical lines
            horiz = false;
            factor*=-1.0f;
        }
        else
            return;//do nothing for 0 factor
            
        final float otherFactor = 1.0f-factor;
        
        for(int i = 0; i < (horiz?img.getHeight():img.getWidth());++i) {
            int lastVal = img.getRGB(i,0);
            for (int j = 1; j < (horiz?img.getWidth():img.getHeight());++j) {
                int thisVal = horiz?img.getRGB(j,i):img.getRGB(i,j);
                int lastR = (lastVal >> 16) & 0xff;
                int thisR = (thisVal >> 16) & 0xff;
                int lastG = (lastVal >> 8) & 0xff;
                int thisG = (thisVal >> 8) & 0xff;
                int lastB = lastVal & 0xff;
                int thisB = thisVal & 0xff;
                
               
                //apply for next time around loop
                lastVal = thisVal;
                
                //change individual "last" components to derivative
                lastR = thisR-lastR;
                lastG = thisG-lastG;
                lastB = thisB-lastB;
                //clamp to 0
                lastR = lastR < 0 ? 0 : lastR; 
                lastG = lastG < 0 ? 0 : lastG;
                lastB = lastB < 0 ? 0 : lastB;
                
                thisR = ((int)(lastR*factor+thisR*otherFactor))&0xff;
                thisG = ((int)(lastG*factor+thisG*otherFactor))&0xff;
                thisB = ((int)(lastB*factor+thisB*otherFactor))&0xff;
                
                if (horiz)
                    img.setRGB(j,i,(thisR << 16)+(thisG << 8)+thisB);
                else
                    img.setRGB(i,j,(thisR << 16)+(thisG << 8)+thisB);
                
            }
        }
    }
    
}
